import curses
import re
from editor import Editor


nonwords = ' .,;-_'

def main():
    running = True
    
    import settings

    default_note = settings.args['default_note']
    editor = Editor(default_note)

    end_state = None

    while running:
        try:
            editor.draw_screen()
            cy, cx = editor.cursor

            c = editor.screen.getch()
            handled_key = False
            for key, shortcut in settings.shortcuts:
                if c == ord(key)-64:
                    if shortcut == 'quit':
                        if editor.altered:
                            result = editor.dialog('Save before quitting? (Y or N)', yesno=True)
                            
                            if result is None: 
                                pass
                            elif result:
                                editor.save_note(editor.current_note)
                                running = False
                            else: running = False
                        else: running = False
                        
                    elif shortcut == 'save':
                        editor.save_note(editor.current_note)
                        if editor.current_note == '**settings**':
                            settings = reload(settings)
                        
                    elif shortcut == 'goto note':
                        result = editor.dialog('Enter the name of the note to load (^C to cancel):')
                        if result:
                            editor.load_note(result)
                            
                    elif shortcut == 'forward':
                        editor.forward()
                        
                    elif shortcut == 'back':
                        editor.back()
                        
                    elif shortcut == 'new note':
                        editor.load_note(None)
                        
                    elif shortcut == 'settings':
                        editor.load_note('**settings**')

                    elif shortcut == 'find':
                        result = editor.dialog('Enter regex to search (^C to cancel):', editor.last_search)
                        if result:
                            editor.find_next(result)
                        
                    elif shortcut == 'paste':
                        if editor.cuts:
                            editor.alter()

                            if cy == len(editor.buffer): editor.buffer.append('')
                            editor.buffer = editor.buffer[:cy] + editor.cuts + editor.buffer[cy:]
                            editor.correct_cursor(cy+len(editor.cuts), cx)
                            editor.cutting = False
                    
                    elif shortcut == 'cut':
                        if cy < len(editor.buffer):
                            editor.alter()

                            if editor.cutting:
                                editor.cuts += [editor.buffer[cy]]
                            else:
                                editor.cuts = [editor.buffer[cy]]
                            
                            editor.buffer = editor.buffer[:cy] + (editor.buffer[cy+1:] if cy < len(editor.buffer)-1 else [])
                            editor.correct_cursor(cy, 0)
                            editor.cutting = True
                        
                    handled_key = True
            if not handled_key:
                if c == curses.KEY_UP:
                    editor.correct_cursor(cy-1, min(cx, len(editor.buffer[cy-1]) if 0 <= cy-1 < len(editor.buffer) else 0))
                    editor.cutting = False
                elif c == curses.KEY_DOWN:
                    editor.correct_cursor(cy+1, min(cx, len(editor.buffer[cy+1]) if 0 < cy+1 < len(editor.buffer) else 0))
                    editor.cutting = False
                elif c == curses.KEY_LEFT:
                    editor.correct_cursor(cy, cx-1)
                elif c == curses.KEY_RIGHT:
                    editor.correct_cursor(cy, cx+1)
                elif c == curses.KEY_LEFT + 279:
                    # ctrl left
                    first = True
                    while first or not (cx==0 or (cy < len(editor.buffer) and len(editor.buffer[cy])>0 and editor.buffer[cy][cx-1] in nonwords 
                        and (cx >= len(editor.buffer[cy]) or not editor.buffer[cy][cx] in nonwords))):
                        first = False
                        editor.correct_cursor(cy, cx-1)
                        cy, cx = editor.cursor
                elif c == curses.KEY_RIGHT + 293:
                    # ctrl right
                    first = True
                    while first or not (cx==0 or (cy < len(editor.buffer) and len(editor.buffer[cy])>0 and editor.buffer[cy][cx-1] in nonwords 
                        and (cx >= len(editor.buffer[cy]) or not editor.buffer[cy][cx] in nonwords))):
                        first = False
                        editor.correct_cursor(cy, cx+1)
                        cy, cx = editor.cursor
                elif c == ord('\n'):
                    follow_link = False
                    for pos, text in editor.links:
                        if pos <= cx < pos+len(text):
                            follow_link = text[1:-1]
                            current_note = follow_link
                            editor.load_note(current_note)
                    if not follow_link:
                        editor.alter()
                        if cy == len(editor.buffer):
                            editor.buffer.append('')
                        else:
                            editor.buffer = (editor.buffer[:cy] + [editor.buffer[cy][:cx]] + 
                                             [editor.buffer[cy][cx:]] + editor.buffer[cy+1:])
                        editor.correct_cursor(cy+1, 0)
                elif c == curses.KEY_BACKSPACE:
                    editor.correct_cursor(cy, cx-1)
                    editor.del_char(cy, cx-1)
                elif c == curses.KEY_DC:
                    editor.alter()
                    editor.del_char(cy, cx)
                elif c == ord('\t'):
                    # tab
                    tab = ' '*settings.args['tab_width']
                    if cy == len(editor.buffer): 
                        editor.buffer.append(tab)
                    else:
                        editor.buffer[cy] = tab + editor.buffer[cy]
                    editor.correct_cursor(cy, cx+len(tab))
                elif c == 353:
                    # shift+tab
                    tab = ' '*settings.args['tab_width']
                    if cy < len(editor.buffer) and editor.buffer[cy].startswith(tab):
                        editor.buffer[cy] = editor.buffer[cy][len(tab):]
                        editor.correct_cursor(cy, cx-len(tab))
                elif c == curses.KEY_HOME:
                    if cx == 0 and cy < len(editor.buffer) and editor.buffer[cy]:
                        stripped = editor.buffer[cy].lstrip(' ')
                        cx = len(editor.buffer[cy]) - len(stripped)
                        editor.correct_cursor(cy, cx)
                    else:
                        editor.correct_cursor(cy, 0)
                elif c == curses.KEY_END:
                    if cy < len(editor.buffer):
                        editor.correct_cursor(cy, len(editor.buffer[cy]))
                elif c == curses.KEY_PPAGE:
                    editor.correct_cursor(cy-editor.height, 0)
                elif c == curses.KEY_NPAGE:
                    editor.correct_cursor(cy+editor.height, 0)
                # TODO: c<255? not all of those are good characters
                elif 0 < c < 255:
                    editor.alter()
                    if cy > len(editor.buffer)-1: editor.buffer += ['']
                    editor.buffer[cy] = editor.buffer[cy][:cx] + chr(c) + editor.buffer[cy][cx:]
                    editor.correct_cursor(cy, cx+1)
                    
                #editor.status = str(c)

        except KeyboardInterrupt:
            running = False
            
        except Exception as e:
            #raise
            editor.status = 'exception: %s' % e
            end_state = editor.status
            #running = False
        
    editor.end_app()
    if end_state: print end_state

if __name__ == '__main__':
    main()



import argparse
import ConfigParser
import os
import sys



shortcuts = [
             ('O', 'save'),
             ('X', 'quit'),
             ('N', 'new note'),
             ('G', 'goto note'),
             ('K', 'cut'),
             ('U', 'paste'),
             ('B', 'back'),
             ('F', 'forward'),
             ('W', 'find'),
             ('T', 'settings'),
             ]

def make_dir_if_not_exists(path):
    if not os.path.exists(path): os.makedirs(path)

home_dir = os.path.expanduser('~/.nanote/')
make_dir_if_not_exists(home_dir)
config_file_path = os.path.join(home_dir, 'settings')

defaults = {
            'path': os.path.join(home_dir, 'notes'),
            'default_note': '',
            'tab_width': '4',
}

config = ConfigParser.SafeConfigParser()
if not config.has_section('nanote'): config.add_section('nanote')
for key, value in defaults.items():
    config.set('nanote', key, value)
config.read(config_file_path)
if not os.path.exists(config_file_path): 
    with open(config_file_path, 'w') as output_file:
        config.write(output_file)

# get note search paths
# TODO: parse sys.argv arguments
args = {}
for arg in ('path', 'default_note', 'tab_width'):
    args[arg] = config.get('nanote', arg)
        
if len(sys.argv) > 1: args['default_note'] = sys.argv[1]
args['tab_width'] = int(args['tab_width'])

NOTE_SEARCH_PATHS = '.:' + args['path']

NOTE_SEARCH_PATHS = [os.path.expanduser(p) for p in NOTE_SEARCH_PATHS.split(':')]


for path in NOTE_SEARCH_PATHS:
    make_dir_if_not_exists(path)

def find_note(note_name):
    # given a note name (i.e. school:homework), search {path}/school/homework for all 
    # paths on the note search path
    if note_name == '**settings**': return config_file_path
    note_name = note_name.replace(':', '/')
    note_paths = [os.path.join(path, note_name) for path in NOTE_SEARCH_PATHS]
    for path in note_paths:
        if os.path.exists(path): return path
    return None
    
def default_note_path(note_name):
    note_name = note_name.replace(':', '/')
    return os.path.join(NOTE_SEARCH_PATHS[-1], note_name)



from setuptools import setup
from __init__ import VERSION


setup(name='nanote',
      version=VERSION,
      description='nanote terminal note-taking software',
      author='Ben Morris',
      author_email='ben@bendmorris.com',
      url='https://github.com/hourchallenge/nanote',
      packages=['nanote'],
      package_dir={
                'nanote':''
                },
      entry_points={
        'console_scripts': [
            'nanote = nanote.nanote:main',
        ],
      },
      )



import curses
import settings as settings_module
import os
import re
from __init__ import VERSION


tab_symbols = ['*', 'o', '>', '-', '+']

class Editor:
    def __init__(self, start_note=None, debug=False):
        import settings
        self.settings = settings
        
        self.start_app()
        
        self.cursor = (0,0)
        self.pad_position = (0,0)
        self.status = ''
        self.last_search = ''
        self.altered = False
        self.history = [None]
        self.history_position = 0
        self.links = []
        self.cutting = False
        self.cuts = []
        
        self.load_note(start_note)
        self.current_note = start_note if start_note else 'untitled'
        
        if not start_note and debug:
            self.buffer = ['This is a test note.', '', 
                           'You can link to other notes like this: [school:homework].', 
                           'Put the cursor over a link and press ENTER to follow it.',
                           'Surround text in asterisks to *bold it*; underscores _underline text_.',
                           '',
                           'Make *bulleted lists* with indentation and asterisks, like this:',
                           '* one', '    * two', '        * three', '            * four',
                           ]


    def start_app(self):
        # start the application
        self.screen = curses.initscr()
        curses.noecho()
        curses.cbreak()
        self.screen.keypad(1)

    def end_app(self):
        # end the application
        curses.nocbreak()
        self.screen.keypad(0)
        curses.echo()
        curses.endwin()
        
    def draw_screen(self):
        settings = self.settings
        
        self.screen.refresh()
        
        screen_size = self.screen.getmaxyx()
        self.height, self.width = screen_size
        height, width = self.height, self.width
        cy, cx = self.cursor
        py, px = self.pad_position

        self.buffer_pad = curses.newwin(height-4, width, 1, 0)
        self.title_win = curses.newwin(1, width, 0, 0)
        self.shortcut_win = curses.newwin(3, width, height-3, 0)

        columns = int((len(settings.shortcuts)+.5)/2)
        column_width = width / (columns + 1)
        for n, shortcut in enumerate(settings.shortcuts):
            x = (n/2) * column_width
            y = 1 if not n%2 else 2
            self.shortcut_win.addstr(y, x, '^' + shortcut[0], curses.A_REVERSE)
            self.shortcut_win.addstr(' ' + shortcut[1])
        
        if self.status: status_text = self.status
        else: status_text = ('"%s"  line %s/%s, col %s/%s, chars %s' % 
                             (self.current_note, cy+1, len(self.buffer) + 1, 
                              cx+1, len(self.buffer[cy]) if cy < len(self.buffer) else 0,
                              sum([len(buffer) for buffer in self.buffer]))
                            )
        
        total_gap = width - len(status_text) - 1
        left_gap = total_gap/2
        right_gap = total_gap-left_gap
        self.shortcut_win.addstr(0, 0, ' '*(left_gap) + status_text + ' '*(right_gap), curses.A_REVERSE)
        self.status = ''
            
        self.shortcut_win.noutrefresh()

        note_name = self.current_note if self.current_note else 'untitled'
        title_text = '  nanote %s' % (VERSION)
        note_name_text = str(note_name) + ('*' if self.altered else '')
        total_gap = width - len(note_name_text) - 1
        left_gap = total_gap/2 - len(title_text)
        right_gap = total_gap-left_gap-len(title_text)
        self.title_win.addstr(title_text + ' '*left_gap + note_name_text + ' '*right_gap, curses.A_REVERSE)
        self.title_win.noutrefresh()

        onscreen_range = range(py, 
                               min(py + height-3 + 1, len(self.buffer) - 1)+1)

        for n, i in enumerate(onscreen_range):
            try:
                start_x = None if cx < width-1 else cx-width+1 
                end_x = max(cx, width-1)
                self.buffer_pad.addstr(n, 0, self.buffer[i][start_x:end_x])
            except: pass
        self.links = []
        link_re = re.compile("\[[a-zA-Z\_\-\.\:\/]+\]")
        bold_re = re.compile("\*[^ ^\[^\]^*^_][^\[^\]^*^_]*?\*")
        underline_re = re.compile("\_[^ ^\[^\]^*^_][^\[^\]^*^_]*?\_")
        bullet_re = re.compile('^ *\* .*')
        comment_re = re.compile('\#.*')
        def draw_trimmed_text(n, pos, text, style):
            start_x = None if cx < width-1 else cx-width+1 
            end_x = max(cx, width-1)
            
            try:
                for i, char in enumerate(text):
                    if start_x <= pos+i < end_x:
                        self.buffer_pad.addstr(n, pos+i, char, style)
            except: pass
            
        for n, i in enumerate(onscreen_range):
            for m in link_re.finditer(self.buffer[i]):
                pos = m.start(); text = m.group()
                draw_trimmed_text(n, pos, text, curses.A_REVERSE)
                if i == cy: self.links.append((pos, text))
            for m in bold_re.finditer(self.buffer[i]):
                pos = m.start(); text = m.group()
                draw_trimmed_text(n, pos+1, text[1:-1], curses.A_BOLD)
            for m in underline_re.finditer(self.buffer[i]):
                pos = m.start(); text = m.group()
                draw_trimmed_text(n, pos+1, text[1:-1], curses.A_UNDERLINE)
            for m in bullet_re.finditer(self.buffer[i]):
                pos = m.start(); text = m.group()
                stripped_text = text.lstrip(' ')
                indentation = len(text)-len(stripped_text)
                indent_level = (indentation / settings.args['tab_width']) % len(tab_symbols)
                draw_trimmed_text(n, pos+indentation, tab_symbols[indent_level], curses.A_BOLD)
            for m in comment_re.finditer(self.buffer[i]):
                pos = m.start(); text = m.group()
                draw_trimmed_text(n, pos, text, curses.A_DIM)

        self.buffer_pad.noutrefresh()

        self.screen.move(cy+1-py, min(cx, width-1))
        self.screen.noutrefresh()
        
        curses.doupdate()


    def find_next(self, find):
        # search for next instance of a regex
        self.last_search = find
        cy, cx = self.cursor
        find_re = re.compile(find)
        current_pos = sum([len(self.buffer[y]) if y < cy else 0 for y in range(len(self.buffer))]) + cy + cx
        first_result = None

        def answer(pos):
            for i, line in enumerate(self.buffer):
                this_len = len(line)+1
                if pos >= this_len:
                    pos -= this_len
                else:
                    return self.correct_cursor(i, pos)

        for m in find_re.finditer('\n'.join(self.buffer)):
            pos = m.start()
            if first_result is None: first_result = pos
            if pos > current_pos: return answer(pos)
        if not (first_result is None):
            self.status = 'Search wrapped to beginning of file'
            return answer(first_result)

        self.status = '"%s" not found' % find

        
    def dialog(self, prompt, default_text='', yesno=False):
        running = True
        entered_text = default_text
        
        while running:
            try:
                self.draw_screen()
                
                total_gap = self.width - len(prompt) - 3 - len(entered_text)
                self.shortcut_win.addstr(0, 0, ' %s %s%s' % (prompt, entered_text, ' '*total_gap), curses.A_REVERSE)
                self.screen.move(self.height-3, len(prompt) + 2 + len(entered_text))
                self.shortcut_win.refresh()
                
                c = self.screen.getch()
                
                if yesno:
                    if c in (ord('y'), ord('Y')):
                        return True
                    elif c in (ord('n'), ord('N')):
                        return False
                    
                else:
                    if c == curses.KEY_BACKSPACE:
                        entered_text = entered_text[:-1]
                    elif c == ord('\n'):
                        return entered_text
                    elif c == 27 or c == curses.KEY_DC:
                        entered_text = ''
                    elif 0 < c < 255:
                        if len(entered_text) < 100:
                            entered_text += chr(c)
                
            except KeyboardInterrupt:
                return None

    
    def correct_cursor(self, cy, cx):
        buffer = self.buffer
        py, px = self.pad_position
        if cy < 0: 
            # up too far; move to top
            cy = 0; cx = 0
        if cy >= len(buffer): 
            # down too far; move to bottom
            cy = len(buffer)
        if cx < 0:
            # left too far; move up and to the end of the previous line
            return self.correct_cursor(cy-1, len(buffer[cy-1]) if 0 <= cy-1 < len(buffer) else 0)
        if (cy < len(buffer)) and (cx > len(buffer[cy])): 
            # right too far; move down and to the beginning of the next line
            return self.correct_cursor(cy+1, 0)
        if (cy >= len(buffer)) and cx > 0:
            cy = len(buffer); cx = 0

        while cy < py: py -= 1
        while cy > py + self.height-5: py += 1
        #while cx < px: px -= 1
        #while cx > self.width-1: cx -= 1
        self.cursor = (cy, cx)
        self.pad_position = (py, px)
        
    def load_note(self, note_name, title=None, going_back = False):
        settings = self.settings
        if not title: 
            title = note_name if note_name else 'untitled'
        
        if self.altered:
            response = self.dialog('Save changes to old note? (Y or N, ^C to cancel)', yesno=True)
            if response is None: return
            self.altered = False
            if response:
                self.save_note(self.current_note)
        try:
            if not note_name:
                self.buffer = []
            else:
                note_path = settings.find_note(note_name)
                if os.path.isdir(note_path):
                    title = '<%s>' % title
                    self.buffer = ['[%s:%s]' % (note_name, file_name) for file_name in sorted(os.listdir(note_path))]
                else:
                    with open(note_path) as note_file:
                        self.buffer = [r.rstrip('\n') for r in note_file.readlines()]    
            if not going_back and not (self.history_position == len(self.history)-1 and self.history[-1] == note_name):
                if note_name in self.history: 
                    self.history.remove(note_name)
                    self.history_position -= 1
                self.history = self.history[:self.history_position+1] + [note_name]
                self.history_position = len(self.history) - 1
        except: 
            self.buffer = []
            self.status = "new note: %s" % note_name
            return self.load_note(None, title=title, going_back=going_back)
            
        self.current_note = title
        self.cursor = (0,0)
        self.altered = False
        
    def save_note(self, note_name):
        settings = self.settings
        
        if note_name == 'untitled': note_name = ''
        
        note_name = self.dialog('Enter the name of the note to save (^C to cancel):', note_name)
        if not note_name: return
        
        note_path = settings.find_note(note_name)
        if not note_path: note_path = settings.default_note_path(note_name)
        directory = '/'.join(note_path.split('/')[:-1])
        settings.make_dir_if_not_exists(directory)
        with open(note_path, 'w') as note_file:
            note_file.write('\n'.join(self.buffer))
        self.status = 'Saved note "%s" to %s' % (note_name, note_path)
        self.altered = False
        self.current_note = note_name
        self.history[self.history_position] = note_name
        
        if note_name == '**settings**':
            self.settings = reload(settings)
        
    def forward(self):
        if self.history_position < len(self.history)-1:
            self.history_position += 1
            self.load_note(self.history[self.history_position], going_back=True)
    
    def back(self):
        if self.history_position > 0:
            self.history_position -= 1
            self.load_note(self.history[self.history_position], going_back=True)

    def alter(self):
        if self.current_note and self.current_note.startswith('<') and self.current_note.endswith('>'): return
        self.altered = True

    def del_char(self, cy, cx):
        self.alter()

        if cx < 0 and cy > 0:
            if cy < len(self.buffer):
                new_buffer = self.buffer[:cy-1]
                new_buffer += [self.buffer[cy-1] + self.buffer[cy]]
                if cy < len(self.buffer)-1: new_buffer += self.buffer[cy+1:]
                self.buffer = new_buffer
        elif cx >= 0:
            self.buffer[cy] = self.buffer[cy][:cx] + self.buffer[cy][cx+1:]

